  /*------------------------------------------------------------------------------------------*\
 *   
 *   Copyright (C) 2016 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
\*------------------------------------------------------------------------------------------*/

#include <linux/linkage.h>
#include <asm/assembler.h>
#include <asm/thread_info.h>

#include <asm/asm-offsets.h>
#include <mach/avm_gic.h>


/* ===   AVM FIQ   === */

	.data
	.global avm_fiq_stacks
	.align
avm_fiq_stacks:
	.long	0

	.text
	.global avm_fiq_handler_begin
	.global avm_fiq_handler_end


/* Initial vorbereitetes Register:
 *  - SP : Stackpointer
 */
ENTRY(avm_fiq_handler)
avm_fiq_handler_begin:

	/* Kompletten Kontext in struct pt_regs sichern, wir wollen rekursiv/reentrant sein ... */
	sub	lr, lr, #4
	str	lr, [sp, #-4]!			@ PC zur Rueckkehr
						@ Ab hier PT_REGS Struktur:
	str	r0, [sp, #-4]!			@ R0_orig
	mrs	r0, spsr
	str	r0, [sp, #-4]			@ CPSR

//      and	r0, r0, #0x1F
//	cmp	r0, #0x11
//	subne	lr, lr, #4

	mov	r0, sp
	sub	r0, r0, #4
	str	lr, [r0, #-4]!			@ PC
	sub	r0, r0, #4			@ LR
	str	sp, [r0, #-4]!			@ SP
	stmdb	r0!, {r1-r12}			@ R1-R12
	mov	sp, r0
	str	r0, [sp, #-4]!			@ R0

	/* Register vorbereiten:
	 *  - R12: GIC CPU-IF Basis
	 *  - R11: FIQ C-Handler
	 *  - R10: User-Ref
	 */
	mov	r8, #THREAD_SIZE
	sub	r8, #0x01
	mov	r9, sp
	bic	r9, r9, r8
	add	r9, r9, r8
	sub	r9, r9, #0x17

	ldr	r10, [r9], #4
	ldr	r11, [r9], #4
	ldr	r12, [r9]	

	/* Ack für den Interrupt, Interrupt-Nr jetzt in R9 */
	ldr	r9, [r12, #0x900]

	/* Test auf gültigen Index */
	tst    r9, #0x80000000

		/* Rote Info LED an
		ldrmi    r8, 2f
		movmi    r9, #0x20
		strmi    r9, [r8]*/

		/* Ruecksprung */
		bmi    1f

	/* Index holen */
	mov    r8, #0x400
	sub    r8, r8, #1
	and    r9, r9, r8			@ untere 10 bit mit 0x3FF herausfiltern


        /* --- Kern der FIQ-Routine ---

	/* Register
	 *  - R12: GIC CPU-IF Basis
	 *  - R11: FIQ C-Handler
	 *  - R10: User-Ref
	 *  - R9 : Interrupt-Nr
	 *  - R8 : hoechste anliegende Interrupt-Prio, maskiert
	 */

	/* FIQ private Register auf Stack sichern */
	stmfd	sp!, {r8-r12}

	/* Argumente für den Handler holen,
	 *  - R0: pt_regs
	 *  - R1: Interrupt-Nr
	 *  - R2: Nutzer-Ref
	 * C-Handler wird ueber R11 angesprungen
	 */

	add	r0, sp, #0x14			@ pt_regs
	mov	r1, r9				@ Interrupt-Nr
	mov	r2, r10				@ User-Ref 

	/* Handler rufen */
	blx r11

	/* Sicherstellen, dass alle
	 *  - Speicherzugriffe
	 *  - Cache-,
	 *  - Branch predictor- und
	 *  - TLB-Verwaltungs-Operationen
	 *  beendet werden, bevor weitere Opcodes
	 *  zur Ausfuehrung kommen.
	 */
	/*dsb*/

	/* FIQ's private Register wieder herstellen */
	ldmfd   sp!, {r8-r12}

        /* ---------------------------- */


	/* EOI */
    	strne	r9, [r12, #0x24]

1:
	/* SPSR fuer Ruecksprung zum vorherigen Modus wieder herstellen */
	ldr	r0, [sp, #64]
	msr	spsr_cxsf, r0

	/* Ruecksprung vom FIQ */
	mov	r0, sp
	add	r0, r0, #4			@ Stack auf pt_regs's R1 stellen
	ldmia	r0!, {r1-r12}			@ R1 - R12 wieder herstellen
	ldr	sp, [r0], #4
	ldr	lr, [r0], #4
	ldmia   sp!, {r0,pc}^			@ Zum Schluss R0 und PC holen, mit Modewechsel ueber SPSR ...

ENDPROC(avm_fiq_handler)

	/* Immediate Literale fuer den FIQ-Handler */
	.align
5:  .long   avm_fiq_stacks			@ Pointer auf den Stack-Bereich
avm_fiq_handler_end:

/* === AVM FIQ END === */


